# -*- coding: utf-8 -*-

"""'prep' subcommand"""

# ----------------------- Imports --------------------------------
import sys
from pathlib import Path
import datetime
from collections import Counter
import csv
import shutil
import os

# My modules
import config
from modules.utils import verbose
from modules.utils import xtriage

#--------------------------Functions------------------------------

def preparation_run(data_dir, args):
    """
    'prep' subcommand
    :param data_dir: data directory
    :param args: args from parser
    :return: log file
    """

    date = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")

    """Folder where we work"""
    work_dir = Path.cwd()

    """Creation of a treatment directory"""
    phenix_dir = work_dir.joinpath('easYPipe')
    phenix_dir.mkdir(exist_ok=True)

    """Creation of a 'prep' folder"""
    prep_dir = phenix_dir.joinpath('1a_prep')
    prep_dir.mkdir(exist_ok=True)

    """Creation of a log file with preparation information"""
    prep_dir.joinpath('logs').mkdir(exist_ok=True)
    nom_prep_log = "prep_easypipe_" + date + ".log"
    prep_log = verbose.log_start(prep_dir, 'logs', nom_prep_log, verbose.easypipe_logo)
    verbose.print_and_write("Treatment directory created: " + str(phenix_dir), prep_log)
    verbose.print_and_write("Creation of '1a_prep' directory", prep_log)

    """List of datasets"""
    verbose.print_and_write(
        verbose.cadre_text(config.motif, config.nb_motifs, "Analysis of datasets"), prep_log)
    jeux_name = []  # liste des dossiers jeux avec leur path
    jeux_name = [j.name for j in data_dir.iterdir() if j.is_dir()]
    jeux_name.sort()
    verbose.print_and_write("Data folder: " + str(data_dir), prep_log)
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
    verbose.print_and_write("List of datasets in data folder (" + str(len(jeux_name)) + " datasets): ", prep_log)
    for j in jeux_name:
        verbose.print_and_write(j, prep_log)

    """List of mtz files"""
    """
    Process from EDNA ESRF :
    grenades_parallelproc, grenades-fastproc => '*_pointless.mtz', 'best_noanom.mtz'
    EDNA_proc => 'ep_*_truncate.mtz'
    autoPROC =>  'ap_*_truncate.mtz'
    XIA2_DIALS => 'di*_free.mtz' ou 'di*_free_*' 
    XDSAPP => '*_F_minus.mtz'
    grenades_codgas, grenades_parallelproc => truncated-unique.mtz
    """
    # tester si mtz EDNA selon le template (car plusieurs mtz dans le dossier)
    mtz_template_proc_data = ['*_pointless.mtz', '*best_noanom.mtz', 'ep_*_truncate.mtz', 'ap_*_truncate.mtz',
                              'di*_free.mtz', 'di*_free_*', '*_F_minus.mtz', 'truncated-unique.mtz']

    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs, "Listing mtz files"),
                            prep_log)
    mtz_files = []  # liste des fichiers mtz à traiter, avec leur path
    # lister TOUS les mtz
    mtz_files_all = data_dir.rglob("*.mtz")
    # pour chaque dossier parent, chercher LE mtz à prendre
    parent_mtz_all = [(mtz.parent, mtz.name) for mtz in mtz_files_all]
    nb_mtz_parent_all = Counter(elem[0] for elem in parent_mtz_all)  # dictionnaire avec {dossier parent : nb de mtz}
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
    for parent, nb_mtz in nb_mtz_parent_all.items():
        if nb_mtz == 1:
            for par, mtz in parent_mtz_all:
                if parent == par:
                    mtz_files.append(parent.joinpath(mtz))
        else:  # cas plusieurs mtz dans un dossier (EDNA), cherche si un mtz correspond à la liste de templates
            for template in mtz_template_proc_data:
                if list(parent.glob(template)):
                    mtz_files.append(list(parent.glob(template))[0])
                    break
    # trier les mtz par ordre alpha
    mtz_files = sorted(mtz_files, key=lambda v: str(v).lower())  # tri par chemin de mtz croissant, gère X pas avant a

    """Copy of datasets to treat in a folder 'processed_datasets'"""
    mtz_copy = []  # liste des mtz copiés pour traitement
    jeux_sans_mtz = []  # liste des jeux sans mtz
    verbose.print_and_write("Creation of the directory: 0_processed_datasets", prep_log)
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
    verbose.print_and_write("Copy of the mtz files from datasets in this directory.", prep_log)
    phenix_dir.joinpath('0_processed_datasets').mkdir(exist_ok=True)
    for j in jeux_name:
        verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
        no_mtz = 0  # numéro du fichier mtz traité (pour le compteur), mise à 0 à chaque jeu
        for mtz in mtz_files:
            if j in mtz.parts and no_mtz == 0:
                # création du datasets (=jeu)
                verbose.print_and_write("Creation of dataset directory: " + j + " in '0_processed_datasets'",
                                              prep_log)
                phenix_dir.joinpath('0_processed_datasets', j).mkdir(exist_ok=True)
                verbose.print_and_write("Creation of 'data' directory for copy of mtz: " + j, prep_log)
                phenix_dir.joinpath('0_processed_datasets', j, 'data').mkdir(exist_ok=True)
                verbose.print_and_write("Creation of 'ligand' directory for future ligand search", prep_log)
                phenix_dir.joinpath('0_processed_datasets', j, 'ligand').mkdir(exist_ok=True)
            if j in mtz.parts:
                no_mtz += 1
                #  et copie du mtz
                proc_name = "mtz" + f'{no_mtz:03d}' + "_".join(mtz.parts[:-1]).split(j,1)[1]    # split à la première occurence du jeu
                proc_dir = phenix_dir.joinpath('0_processed_datasets', j, 'data', proc_name)
                proc_dir.mkdir(exist_ok=True)
                if not (proc_dir.joinpath(mtz.name)).exists():  # évite de recopier en cas de relancement de prep
                    verbose.print_and_write(mtz.name + ": copied in '" + proc_name + "' ...", prep_log)
                    shutil.copyfile(str(mtz), str(proc_dir.joinpath(mtz.name)))
                else:
                    verbose.print_and_write(mtz.name + ": already copied in '" + proc_name + "' ...", prep_log)
                mtz_copy.append(proc_dir.joinpath(mtz.name))
                # copie des fichiers log et de type XDS_ACSII.HKL et mtz contenant 'unmerged' ou 'multirecord'
                # situés dans le même dossier que le mtz
                log_data_files = (file for file in mtz.parent.iterdir() if file.is_file() and (file.name.endswith('.log')
                            or file.name == 'CORRECT.LP' or '.HKL' in file.name
                            or (('unmerged' in file.name or 'multirecord' in file.name) and '.mtz' in file.name)))
                verbose.print_and_write("Copy of log files or files useful for deposition:", prep_log)
                for file in log_data_files:
                    if not proc_dir.joinpath(file.name).exists():
                        verbose.print_and_write("\t" + file.name, prep_log)
                        shutil.copyfile(str(file), str(proc_dir.joinpath(file.name)))
                    else:
                        verbose.print_and_write("\t" + file.name + ": already copied ...", prep_log)
        if no_mtz == 0:  # cas pas de mtz pour le jeu
            jeux_sans_mtz.append(j)

    """Launch Xtriage on each mtz file to get informations"""
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs,
                                                           "Getting informations from mtz files with Xtriage"),
                                  prep_log)
    # Xtriage
    # Faire tourner xtriage
    verbose.print_and_write("Xtriage launched ...", prep_log)
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
    for mtz in mtz_copy:
        nom_xtriage = mtz.parent.joinpath("xtriage_" + mtz.stem + ".log")
        xtriage_verbose = mtz.parent.joinpath("xtriage-verbose.log")
        if not nom_xtriage.exists(): # TODO à voir: xtriage_verbose changé pour nom_xtriage => ??
            verbose.print_and_write("Xtriage launched for: " + "/".join(mtz.parts[-4:]), prep_log)
            xtriage.xtriage_mtz(mtz, xtriage_verbose, nom_xtriage)
        else:
            verbose.print_and_write("Xtriage already launched for: " + "/".join(mtz.parts[-4:]), prep_log)

    """Write informations in a csv file: space group, resolution and completness for each mtz file"""
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs,
                                                           "Writing informations from mtz files in a csv file"),
                                  prep_log)
    verbose.print_and_write("csv file created: mtz_to_treat_ALL.csv in directory " + prep_dir.name, prep_log)
    verbose.print_and_write("dataset/mtz dir./mtz: resolution / completeness / space group / unit cell", prep_log)
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
    mtz_treat_csv = prep_dir.joinpath('mtz_to_treat_ALL.csv')
    header = ["dataset", "mtz nb", "process name", "mtz file", "resolution(A)", "completeness(%)",
              "space group", "unit cell"]
    with open(mtz_treat_csv, 'w', newline='') as csvfile:
        csv_writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(header)  # header
        for j in jeux_name:
            for mtz in mtz_copy:
                if j in mtz.parts:
                    for fi in mtz.parent.iterdir():
                        # Récupération des infos dans les log Xtriage
                        if fi.is_file and fi.name.startswith("xtriage_"):
                            resolution = xtriage.reso_in_xtriage_log(fi)
                            completeness = xtriage.completeness_in_xtriage_log(fi)
                            space_group = xtriage.sg_in_xtriage_log(fi)
                            unit_cell = xtriage.unitcell_in_xtriage_log(fi)
                            break
                    verbose.print_and_write(
                        j + "/" + mtz.parent.name + "/" + mtz.name + ": " + str(resolution) + "A / "
                        + str(completeness) + "% / " + space_group + " / " + unit_cell, prep_log)
                    csv_writer.writerow([j, mtz.parent.name.split('_')[0], '_'.join(mtz.parent.name.split('_')[1:]),
                                         mtz.name, resolution, completeness, space_group, unit_cell])

    """Create csv file for ligands smiles"""
    verbose.print_and_write(
        verbose.cadre_text(config.motif, config.nb_motifs, "Creating a csv file for ligand search"),
        prep_log)
    verbose.print_and_write("Creation of the directory: /1c_ligands", prep_log)
    lig_dir = phenix_dir.joinpath('1c_ligands')
    lig_dir.mkdir(exist_ok=True)
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
    lig_file = lig_dir.joinpath('ligands_for_datasets.csv')
    verbose.print_and_write("Creation of a csv file in this directory: " + lig_file.name, prep_log)
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
    with open(lig_file, 'w', newline='') as csvfile:
        csv_writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(["dataset", "ligand name", "ligand smiles"])
        for j in jeux_name:
            if j not in jeux_sans_mtz:  # cas 'pas de mtz' géré (ne copie pas le dossier dans processed_datasets)
                csv_writer.writerow([j])
    verbose.print_and_write(
        "If you want Phenix to try to place ligands from smiles, fill in the fields 'ligand name' "
        "and 'ligand smiles',\n"
        "then save the csv file somewhere else if you don't want to overwrite it in case you launch "
        "'prep' subcommand again ...", prep_log)

    """Summary and next steps ..."""
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs, "SUMMARY"), prep_log)
    verbose.print_and_write("List of copied datasets (" + str(len(jeux_name) - len(jeux_sans_mtz)) + "): ",
                                  prep_log)
    for j in jeux_name:
        if j not in jeux_sans_mtz:
            verbose.print_and_write(j, prep_log)
    if jeux_sans_mtz:
        verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
        verbose.print_and_write("List of datasets without mtz file (" + str(len(jeux_sans_mtz)) + "): ", prep_log)
        for j in jeux_sans_mtz:
            verbose.print_and_write(j, prep_log)
        # écrire les jeux sans mtz dans un fichier csv
        datasets_no_mtz = prep_dir.joinpath("datasets_without_mtz.csv")
        with open(datasets_no_mtz, 'w', newline='') as csvfile_no_mtz:
            csv_writer = csv.writer(csvfile_no_mtz, quoting=csv.QUOTE_MINIMAL)
            csv_writer.writerow(['dataset'])  # header
            for j in jeux_sans_mtz:
                csv_writer.writerow([j])
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), prep_log)
    if not args.subcommands == 'auto':
        verbose.print_and_write("Next steps with subcommands:\n", prep_log)
        verbose.print_and_write(
            "\t- reindex (optional): allow to change space group according to a reference mtz when possible,\n"
            "\t                      in case some datasets have mtz that should be in higher symmetry space group"
            "\n Can be launched after a first turn of 'launch' subcommand. "
            "Just launch again 'launch' subcommand.\n", prep_log)
        verbose.print_and_write(
            "\t- ligands: to generate pdb and cif of ligands in corresponding processed dataset folders,\n"
            "\t       mandatory if you want Phenix to search ligand ... else optional.\n", prep_log)
        verbose.print_and_write(
            "First, you have to fill in the fields 'ligand name' and 'ligand smiles' of /1c_ligands/"
            + lig_file.name + " csv file, then save the csv file somewhere else or with another name"
                              " if you don't want to overwrite it in case you launch 'prep' subcommand again ...\n",
            prep_log)
        verbose.print_and_write(
            "\t- launch: to launch Phenix.ligand_pipeline on each mtz file, according to options\n", prep_log)
        verbose.print_and_write(
            "\nFor more help on each subcommand " + Path(sys.argv[0]).name + " 'data folder' 'subcommand' -h", prep_log)
    else:
        verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs, "Next step in automatic mode: subcommand 'reindex' ..."), prep_log)
    os.chdir(work_dir)
    return prep_log